#ifndef Progress_h
#define Progress_h

#include "HProgressBar.h"
#include "ProgressMinion.h"
#include <QtCore/QObject>
#include <QtCore/QTime>
#include <QtCore/QString>

//
// This class controls two progress bars: total progress and item progress.
// It is for large-scale operations which are composed of a number of smaller
// operations, called items.  The total progress indicates progress for the
// whole large-scale operation.  The item progress indicates progress for
// each smaller operation.
//
// The large-scale operation is divided into a total number of steps.  The units
// of the steps can be meaningful (e.g. bytes) or totally arbitary.  Each item
// in the operation also is divided into a number of steps, but the units for
// item steps need not have any relation to the units of steps of the
// large-scale operation.  For instance, a large-scale operation may have 1000
// steps.  And each item may have 100 steps, or 1000 steps, or 10000 steps, or
// any other number.  As progress is made in an item, the item progress is
// scaled to its part of the total operation.
//
// To start a large-scale operation, call startTotal, specifying as activity
// the name of the large-scale operation, and as totalSteps the number of
// steps in the large-scale operation.
//
// To start each smaller operation, call startItem, specifying as the name of
// the item, as totalSteps the number of total steps for the item, and as
// itemSteps the number of item steps for the item.
//
// The progress for each item is updated by calling setItemValue.  Its value is
// the amount of progress, measured in itemSteps for the current item.
//
// Sometimes a block of total steps may be reserved for a number of sub-items
// (acts), frequently determined dynamically.  In this case, an item is divided
// into a number of acts (although the degenerate case where an item is divided
// into only single act is useful, too).  The number of acts need have NO
// RELATION to the number of total steps or the number of item steps for each
// item in the block.
//
// To use acts, first call setupItemActs.  This does not actually start any
// item; it just specifies number of total steps in the block reserved for
// the items, and the number of acts.  Then, for each item, call
// startItemAct, specifying the name of the activity, the number of item
// steps for the act, and the number of the act (which should be one or more
// greater than the value passed in the previous call).  When all acts have
// been completed, call endItemActs.
//
// When the progress bar no longer is being used to display any progress,
// call reset.
//
// For example:
//    obj.startTotal("long operation", 100 + 50 + 200);
//    obj.startItem("part 1", 100, 5000);
//    for (i = 0; i < 5000; i++) {
//       ...
//       obj.setItemValue(i);
//    }
//    obj.startItem("part 2", 50, 3000);
//    for (i = 0; i < 3000; i++) {
//       ...
//       obj.setItemValue(i);
//    }
//    obj.setupItemActs(200, numActs * 5);
//    for (act = 0; act < numActs; act += 5) {
//       obj.startItemAct(QString("part %1").arg(act+3), 1000, act);
//       for (i = 0; i < 1000; i++) {
//          ...
//          obj.setItemValue(i);
//       }
//    }
//    obj.endItemActs();
//    obj.reset();
// And the above example shows progress broken down as follows 
// (with numActs = 3):
//    total = |0.....99|100....149|150...216|217...282|283...349|
//    acts  =                     |0...... 4|5.......9|10.....14|
//    items = |0...4999|0.....2999|0.....999|0.....999|0.....999| 
// The total progress bar tracks the first row.  The item progress bar
// tracks the last row, and thus restarts for each item.  No progress bar
// tracks the middle row; it's shown only so that the scaling relationships
// between the various numbers: total steps, acts, item steps.
//
// Sometimes, a subset of progress items may be performed in parallel.  In this
// case, the item progress bar shows progress for the whole parallel subset, and
// a tooltip can be brought up by the user to get detailed information.  This is
// supported only for acts.  Once a set of acts is setup, start a parallel
// subset of them by calling setItemActsThreads.  It specifies the range of acts
// consumed by the parallel subset and the number of parllel activities actually
// happening (which need not be the same as the number of acts).  Start
// individual parallel threads within the act thread set using
// startItemActThread, and end them with endItemActThread.  End the whole
// parallel subset with endItemActsThreads.  When updating progres in individual
// threads, instead of using setItemValue, it's necessary to call
// setItemActsThreadValue instead of setItemValue, because it's necessary to
// indicate which thread to which the progress value applies.
//
// For example:
//    obj.startTotal("long operation", 100 + 200);
//    obj.startItem("part 1", 100, 5000);
//    for (i = 0; i < 5000; i++) {
//       ...
//       obj.setItemValue(i);
//    }
//    obj.setupItemActs(200, 20);
//    for (subset = 0; subset < 2; subset++) {
//       obj.setupItemActsThreads(QString("subset %1").arg(subset),
//                                subset*10, (subset+1)*10, 7);
//       for (thread = 0; thread < 7; thread++) {
//          startItemActThread(QString("subset %1 thread %2")
//                                .arg(subset).arg(thread),
//                             thread, 1000);
//          ... start threads ...
//       }
//       EACH THREAD DOES:
//          for (i = 0; i < 1000; i++) {
//             ...
//             obj.setItemActsThreadValue(thread, i);
//          }
//       END EACH THREAD
//       for (thread = 0; thread < 7; thread++) {
//          ... join threads ...
//          endItemActThread(thread);
//       }
//       obj.endItemActsThreads();
//    }
//    obj.endItemActs();
//    obj.reset();
//
// Sometimes, it may be useful just to use the total progress bar and leave
// the item progress bar disabled.  In that case, and only in that case, 
// setTotalValue can be used.  For example:
//    obj.startTotal("long operation", 100);
//    for (i = 0; i < 100; i++) {
//       ...
//       obj.setTotalValue(i);
//    done
//    obj.reset();
//

namespace Progress {

   class Object : public QObject {
      Q_OBJECT
   public:
      explicit Object(HProgressBar*  totalProgress,
                      HProgressBar*  itemProgress);
      ~Object();
   public:
      void startTotal             (QString  activity,
                                   Steps    totalSteps);
      void startItem              (QString  activity,
                                   Steps    totalSteps,
                                   Steps    itemSteps = 1);
      void changeItemActivity     (QString  activity);
      // Interfaces for dynamically-determined number of "acts" within an item.
      void setupItemActs          (Steps    totalSteps,
                                   Steps    numActs);
      void changeItemActs         (Steps    numActs);
      void startItemAct           (QString  activity,
                                   Steps    itemSteps,
                                   Steps    act);
      void endItemActs            ();
      // Interfaces for blocks of acts performed in parallel.
      void setupItemActsThreads   (QString  activity,
                                   Steps    firstAct,
                                   Steps    tailAct,
                                   Threads  numThreads,
                                   Steps    itemSteps);
      void startItemActThread     (QString  activity,
                                   Threads  thread,
                                   Steps    itemSteps);
      void setItemActsThreadValue (Threads  thread,
                                   Steps    value);
      void endItemActThread       (Threads  thread);
      void retryItemActThread     (Threads  thread);
      void endItemActsThreads     ();
      // Interfaces for changing the progress bar position.
      void setItemValue           (Steps    value);
      void setTotalValue          (Steps    value);
      void reset                  ();
   signals:
      void startTotalSig         (QString              activity,
                                  Steps                totalTail);
      void startItemSig          (QString              activity,
                                  Steps                totalValue,
                                  Steps                itemTail);
      void changeItemActivitySig (QString              activity);
      void changeItemActsSig     (Steps                totalValue);
      void setItemValueSig       (Steps                totalValue,
                                  Steps                value);
      void endItemActsSig        (Steps                totalValue);
      void setTotalValueSig      (Steps                value);
      void updateThreadsStateSig (const ThreadsState&  threads);
      void resetSig              ();
   protected:
      void updateGui ();
   protected:
      Minion*      minion;
      bool         singleThread;
      static bool  registeredTypes;
      QTime        lastSetValue;
      QTime        lastGuiUpdate;
   protected:
      Steps    totalValue;      // measured in total steps
      Steps    totalTail;       // measured in total steps

      Steps    itemTotalStart;  // measured in total steps
      Steps    itemTotalTail;   // measured in total steps

      Steps    itemValue;       // measured in item steps
      Steps    itemTail;        // measured in item steps

      Steps    actsTotalSteps;  // measured in total steps
      Steps    actsTotalTail;   // measured in total steps
      Steps    numActs;         // measured in acts
      Steps    prevAct;         // meausred in acts

      Steps    threadsFirstAct; // measured in acts
      Steps    threadsNumActs;  // measured in acts
      ThreadsState  threads;
   };

   // This class is just a handy way to declare a single object that, when
   // destroyed (e.g. on exit from a routine) will reset the Progress.
   class User {
   public:
      explicit User(Object*&       pg,
                    HProgressBar*  totalProgress,
                    HProgressBar*  itemProgress) 
         : pg(pg)
      {
         pg = new Object(totalProgress, itemProgress);
      }
      ~User() 
      { 
         delete pg;
         pg = NULL;
      }
   protected:
      Object*&  pg;
   };

};

#endif
